//
// SQL_db.cc
//
// SQL implementation, written by Zoran Constantinescu <zoran@idi.ntnu.no>
//
#if RELEASE
static char RCSid[] = "$Id: SQL_db.cc $";
#endif

#include <Configuration.h>
#include "SQL_db.h"
#include "HashWord.h"

#include <stdlib.h>
#include <stdio.h>
#include <sys/timeb.h>

#include <stream.h>

extern Configuration config;
extern int           debug;

FILE *SQL_db::SQL_log = NULL;
int   SQL_db::SQL_n = 0;
int   SQL_db::SQL_id = 0;

int   SQL_db::SQL_debug;
int   SQL_db::isWordsHash = 0;
List *SQL_db::lWordsHash[26];

SQL_db::SQL_db ()
{
  if (! SQL_n)
  {
    SQL_log = fopen (config["SQL_log_file"], "a"); 

    fprintf (SQL_log, "\n\n");
  }
  SQL_n++;

  SQL_debug = config.Value ("SQL_debug");

  // no hash, will be built when necessary
  //isWordsHash = 0;
}


SQL_db::~SQL_db ()
{
  SQL_n--;
  if (! SQL_n)
    fclose (SQL_log);

/*zrn Not anymore! the list is a static member (for speedup :)
  if (isWordsHash)
    for (int i = 0; i < 26; i++)
      if (lWordsHash[i]) {
        lWordsHash[i]->Release ();
        delete lWordsHash[i];
      }
*/
}


/* filename = user:pswd@host:database */
int
SQL_db::OpenReadWrite (char *filename, int mode)
{
  return OpenRead (filename);
}

int
SQL_db::OpenRead (char *filename)
{
  char *tmp;
  String user, pswd, host, db;

  filename = config["SQL_dbase"];

  tmp = filename;
  while ((tmp[0] != ':') && (tmp[0] != '\0'))
    tmp++;
  if (tmp[0] == '\0') return NOTOK;
  user.append (filename, tmp - filename);

  tmp++;
  filename = tmp;
  while ((tmp[0] != '@') && (tmp[0] != '\0'))
    tmp++;
  if (tmp[0] == '\0') return NOTOK;
  pswd.append (filename, tmp - filename);

  tmp++;
  filename = tmp;
  while ((tmp[0] != ':') && (tmp[0] != '\0'))
    tmp++;
  if (tmp[0] == '\0') return NOTOK;
  host.append (filename, tmp - filename);

  tmp++;
  filename = tmp;
  while (tmp[0] != '\0')
    tmp++;
  db.append (filename, tmp - filename);

  Connect (host, db, user, pswd);

  return OK;
}


int
SQL_db::Close ()
{
  Disconnect ();

  return OK;
}


int
SQL_db::LogTime (int err, char *sFunction, timeb tt0)
{
timeb tt1;
ftime (&tt1);
int nsecs = 1000 * tt1.time + tt1.millitm -
            1000 * tt0.time - tt0.millitm;

  if (SQL_debug > 3)
    fprintf (SQL_log, "  - %s = %d.%03d sec\n", sFunction, nsecs/1000, nsecs%1000);
  return err;
}


void
SQL_db::LogError (String &s)
{
  if (SQL_debug > 0)
    fprintf (SQL_log, "ERR:%d: %s", SQL_id, s.get ());
}


void
SQL_db::LogLine (String &s, int level)
{
  if (SQL_debug >= level)
    fprintf (SQL_log, "LOG:%d: %s", SQL_id, s.get ());
}


/*****************************************************************/
/*  Add a URL to the database, it can be new or not              */
/*****************************************************************/
int
SQL_db::AddURL (String &sDocURL, int &nDocID, String &sDocRefs)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_update;

  if (sDocURL.length () == 0)
  {
    String s;
    s << "AddURL() -- failed -- "
      << "doc_url='" << sDocURL << "' "
      << "doc_ref+='" << sDocRefs << "';"
      << "error: doc_url length zero?\n";
    LogLine (s, 2);

    return LogTime (NOTOK, "AddURL", t0);
  }

  String s;
  s << "AddURL("
    << "doc_url='" << sDocURL << "' "
    << "doc_ref+='" << sDocRefs << "') ...\n";
  LogLine (s, 2);

  // find out the ID of the document ...
  // (if it doesn't exists it will be added to URLS and URLINFO)
  if (GetDocID (sDocURL, nDocID) == NOTOK)
    return LogTime (NOTOK, "AddURL", t0);

  // get rid of '-es
  QuoteString (sDocRefs);

  // ... add references from other docums.
  s_update << "update URLINFO set doc_refs=concat(doc_refs,' ',"
           << "'" << sDocRefs << "') where doc_id=" << nDocID;
  
  if (! (err = ExecQuery (s_update))) 
  {
    String s;
    s << "AddURL() -- update URLINFO -- "
      << "doc_id=" << nDocID << " "
      << "doc_url='" << sDocURL << "' "
      << "doc_ref+='" << sDocRefs << "'.\n";
    LogLine (s, 2);
  }  
  else
  {
    String s;
    s << "AddURL() -- failed -- "
      << "doc_url='" << sDocURL << "' "
      << "doc_ref+='" << sDocRefs << "'; "
      << "error: '" << Error () << "'.\n"; 
    LogError (s);

    return LogTime (NOTOK, "AddURL", t0);
  }

  return LogTime (OK, "AddURL", t0);
}


/*****************************************************************/
/*  Get a docum. with DOC_STAT_NEW                               */
/*****************************************************************/
int
SQL_db::GetNewURL (String &sDocURL, int &nDocID, int &nServerID)
{
  timeb t0;
  ftime (&t0);

  String sLimitDom;

  return GetNewURL (sDocURL, nDocID, nServerID, sLimitDom);
}


int
SQL_db::GetNewURL (String &sDocURL, int &nDocID, int &nServerID, String &sLimitDom)
{
  timeb t0;
  ftime (&t0);

String sTable ("URLS write");
Lock (sTable);

  int err;
  nDocID = 0;
  int nDocStatus;
  String s_select, s_update;
  int result;
  char **row;

  nDocStatus = DOC_STAT_NEW;
  QuoteString (sLimitDom);

  if (sLimitDom.length () > 0)
    s_select << "select doc_id, doc_url, doc_server_id from URLS where doc_status="
             << nDocStatus << " and doc_url regexp 'http:\\/\\/[^/]*\\."
             << sLimitDom << ".*' limit 1";
  else
    s_select << "select doc_id, doc_url, doc_server_id from URLS where doc_status="
             << nDocStatus << " limit 1";

  if (! (err = ExecQuery (s_select))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      nDocID = (row[0])?atoi (row[0]):0;
      nServerID = (row[2])?atoi (row[2]):0;
      sDocURL = row[1];

      s_update << "update URLS set "
               << "doc_status=" << DOC_STAT_WORK << " "
               << "where doc_id=" << nDocID;

      // set status to WORK
      if (! (err = ExecQuery (s_update))) 
      {
        String s;
        s << "GetNewURL() -- update URLS -- "
          << "doc_id=" << nDocID << " "
          << "server_id=" << nServerID << " "
          << "doc_url='" << sDocURL << "'.\n";
        LogLine (s, 2);

        err = OK;
      }
      else
      {
        String s;
        s << "GetNewURL() -- failed to set STATUS_WORK --\n";
        LogError (s);
 
        err = NOTOK;
      }
    }
    else
    {
      String s;
      s << "GetNewURL() -- no new url --\n";
      LogLine (s, 2);
 
      err = NOTOK;
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetNewURL() -- failed -- "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    err = NOTOK;
  }

UnLock ();

  return LogTime (err, "GetNewURL", t0);
}


/*****************************************************************/
/* Return the URL of the specified document id                   */
/*****************************************************************/
int
SQL_db::GetURL (int nDocID, String &sDocURL)
{
  timeb t0;
  ftime (&t0);

  int err;
  sDocURL = 0;
  String s_select_url;
  int result;
  char **row;

  // ... check for the docum. in URLS ...
  s_select_url << "select doc_url from URLS where doc_id=" << nDocID;

  if (! (err = ExecQuery (s_select_url))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      sDocURL = 0;
      sDocURL.append (row[0]);

      String s;
      s << "GetURL() -- select URLS -- "
        << "doc_id=" << nDocID << " "
        << "doc_url='" << sDocURL << "'.\n";
      LogLine (s, 2);
    }
    else
    {
      String s;
      s << "GetURL() -- no record in URLS -- "
        << "doc_id=" << nDocID << ".\n"; 
      LogLine (s, 2);
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetURL() -- failed -- "
      << "doc_id=" << nDocID << "; "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "GetURL", t0);
  }

  return LogTime (OK, "GetURL", t0);
}


/*****************************************************************/
/*  Return the DOC_ID of the URL                                 */
/*    - create the URL,URLINFO records if necessary              */
/*    - create the link to SERVERS                               */
/*****************************************************************/
int
SQL_db::GetDocID (String &sDocURL, int &nDocID)
{
  timeb t0;
  ftime (&t0);

  int err, isURL = 0, isInfo = 0;
  nDocID = 0;
  int nServID = 0, nDocStatus;
  String s_insert_url, s_select_url,
         s_insert_info, s_select_info,
         sServAddr;
  int result;
  char **row;

  QuoteString (sDocURL);

  // ... check for the docum. in URLS ...
  s_select_url << "select doc_id from URLS where doc_url="
               << "'" << sDocURL << "'";

  if (! (err = ExecQuery (s_select_url))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      nDocID = (row[0])?atoi (row[0]):0;
      //printf ("* in %d", nDocID);
      isURL = 1;

      String s;
      s << "GetDocID(doc_url='" << sDocURL << "') -- select URLS -- "
        << "doc_id=" << nDocID << ".\n";
      LogLine (s, 2);
    }
    else
    {
      isURL = 0;

      String s;
      s << "GetDocID(doc_url='" << sDocURL << "') -- no record in URLS -- \n";
      LogLine (s, 2);
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetDocID(sDocURL='" << sDocURL << "') -- failed1 -- "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "GetDocID", t0);
  }

  // ... if not, then add
  if (! isURL)
  {
    // ... add the server for this docum.
    // eg.   string  http://www.x.ro/ddd/ddd
    //       nStat   00000112222222233333333
    int nStat = 0; // start
    for (int i = 0; i < sDocURL.length (); i++)
    {
      if ((sDocURL[i] == '/') && (nStat == 0))
        nStat = 1; // have first '/'
      else if ((sDocURL[i] == '/') && (nStat == 1))
        nStat = 1; // have another '/'
      else if ((sDocURL[i] != '/') && (nStat == 1))
        nStat = 2; // can copy
      else if ((sDocURL[i] == '/') && (nStat == 2))
        nStat = 3; // end of copy
  
      if (nStat == 2)
        sServAddr << sDocURL[i];
    }
    AddServer (sServAddr, nServID, SERV_STAT_NEW);

    nDocStatus = DOC_STAT_NEW;
    s_insert_url << "insert into URLS (doc_url, doc_status, doc_server_id) "
                 << "values (" << "'" << sDocURL << "',"
                               << nDocStatus << ","
                               << nServID << ")";

    if (! (err = ExecQuery (s_insert_url)))
    {
      nDocID = GetInsertID ();

      String s;
      s << "GetDocID() -- insert URLS -- "
        << "doc_url='" << sDocURL << "' "
        << "doc_id=" << nDocID << ".\n";
      LogLine (s, 2);
    }
    else
    {
      String s;
      s << "GetDocID() -- failed2 -- "
        << "doc_url='" << sDocURL << "'; "
        << "error: '" << Error () << "'.\n";
      LogError (s);

      return LogTime (NOTOK, "GetDocID", t0);
    }
  }
  // check for docum. in URLINFO
  s_select_info << "select doc_id from URLINFO where doc_id="
                << "'" << nDocID << "'";

  s_insert_info << "insert into URLINFO (doc_id) "
                << "values (" << nDocID << ")";

  if (! (err = ExecQuery (s_select_info))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
      isInfo = 1;
    else
    {
      isInfo = 0;    

      String s;
      s << "GetDocID() -- no record in URLINFO -- "
        << "doc_id=" << nDocID << "'.\n";
      LogLine (s, 2);
    }
    FreeResult (result);
  }  
  else
  {
    String s;
    s << "GetDocID() -- failed3 -- "
      << "doc_id=" << nDocID << "; "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "GetDocID", t0);
  }
  // ... if not, then add
  if (! isInfo)
    if (! (err = ExecQuery (s_insert_info)))
    {
      String s;
      s << "GetDocID() -- insert URLINFO -- "
        << "doc_id=" << nDocID << "'.\n";
      LogLine (s, 2);
    }
    else
    {
      String s;
      s << "GetDocID() -- failed4 -- "
        << "doc_url='" << sDocURL << "'; "
        << "error: '" << Error () << "'.\n";
      LogError (s);
  
      return LogTime (NOTOK, "GetDocID", t0);
    }

  return LogTime (OK, "GetDocID", t0);
}


/*****************************************************************/
/* Return all info for the specified document id                 */
/*****************************************************************/
int
SQL_db::GetDocInfo (int nDocID, DocumentRef *doc)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_select;
  int result;
  char **row;

  // ... get the docum. from URLS,URLINFO ...
  s_select << "select doc_url, unix_timestamp(doc_tmodif), "
           << "unix_timestamp(doc_taccess), doc_text, "
           << "doc_title, doc_refs, doc_status, doc_size, "
           << "doc_links, doc_anchors, doc_hopcount "
           << "from URLS,URLINFO where URLS.doc_id=" << nDocID << " "
           << "and URLS.doc_id=URLINFO.doc_id";

  if (! (err = ExecQuery (s_select))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      doc->DocID		(nDocID);

      String sDocURL = doc->DocURL ();
      sDocURL.append (row[0]);
      doc->DocURL (sDocURL);

      doc->DocTime		((time_t)(row[1])?atoi (row[1]):0);
      doc->DocAccessed		((time_t)(row[2])?atoi (row[2]):0);

      String sDocHead = doc->DocHead ();
      sDocHead.append (row[3]);
      doc->DocHead (sDocHead);

      String sDocTitle = doc->DocTitle ();
      sDocTitle.append (row[4]);
      doc->DocTitle (sDocTitle);

      //doc->DocDescriptions	(row[5]);  // List of refs.
      int nDocState		= (row[6])?atoi (row[6]):0;

switch (nDocState)
{
case DOC_STAT_OK:
  doc->DocState (Reference_ok); break;
case DOC_STAT_NEW:
  doc->DocState (Reference_new); break;
case DOC_STAT_WORK:
  doc->DocState (Reference_work); break;
case DOC_STAT_NOT_FOUND:
  doc->DocState (Reference_not_found); break;
case DOC_STAT_NO_HOST:
  doc->DocState (Reference_no_host); break;
case DOC_STAT_NO_SERVER:
  doc->DocState (Reference_no_server); break;
default:
  doc->DocState (Reference_unknown);
}

      doc->DocSize		((row[7])?atoi (row[7]):0);
      doc->DocLinks		((row[8])?atoi (row[8]):0);
      //doc->DocAnchors		(row[9]);  // List of anchors
      doc->DocHopCount		((row[10])?atoi (row[10]):0);

      String s;
      s << "GetDocInfo("
        << "doc_id=" << nDocID << ")\n";
      LogLine (s, 2);
    }
    else
    {
      String s;
      s << "GetDocInfo("
        << "doc_id=" << nDocID << "') => no record!\n";
      LogLine (s, 2);
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetDocInfo("
      << "doc_id=" << nDocID << ") => "
      << "error: '" << Error () << "'!\n";
    LogError (s);

    return LogTime (NOTOK, "GetDocInfo", t0);
  }

  return LogTime (OK, "GetDocInfo", t0);
}


/*****************************************************************/
/*  Update information about a document                          */
/*****************************************************************/
int
SQL_db::AddDocInfo (int nDocID, DocumentRef *doc, int bUpdateURL)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_update1, s_update2, s_update3;
  int result;
  char **row;

  String sDocURL = doc->DocURL ();
  QuoteString (sDocURL);
  String sDocTitle = doc->DocTitle ();
  QuoteString (sDocTitle);
  String sDocHead = doc->DocHead ();
  QuoteString (sDocHead);
  String sDocServerType = doc->DocServerType ();
  QuoteString (sDocServerType);

  int nDocStatus;
  switch (doc->DocState ())
  {
    case Reference_ok:
      nDocStatus = DOC_STAT_OK; break;
    case Reference_new:
      nDocStatus = DOC_STAT_NEW; break;
    case Reference_work:
      nDocStatus = DOC_STAT_WORK; break;
    case Reference_not_found:
      nDocStatus = DOC_STAT_NOT_FOUND; break;
    case Reference_no_host:
      nDocStatus = DOC_STAT_NO_HOST; break;
    case Reference_no_server:
      nDocStatus = DOC_STAT_NO_SERVER; break;
    default:
      nDocStatus = DOC_STAT_UNKNOWN; break;
  }

  String sTable ("URLS write, URLINFO write, SERVERS write");
  Lock (sTable);

  // ... update the docum. in URLS,URLINFO ...
if (bUpdateURL == 1)
  s_update1 << "update URLS set "
            << "doc_url='" << sDocURL << "', "
            << "doc_taccess=from_unixtime(" << doc->DocAccessed () << "), "
            << "doc_status=" << nDocStatus << " "
            << "where doc_id=" << nDocID;
else
  s_update1 << "update URLS set "
            << "doc_taccess=from_unixtime(" << doc->DocAccessed () << "), "
            << "doc_status=" << nDocStatus << " "
            << "where doc_id=" << nDocID;

  s_update2 << "update URLINFO set "
            << "doc_tmodif=from_unixtime(" << doc->DocTime () << "), "
            << "doc_title='" << sDocTitle << "', "
            << "doc_text='" << sDocHead << "', "
            << "doc_size=" << doc->DocSize () << ", " 
            << "doc_links=" << doc->DocLinks () << ", "
          //<< "doc_anchors='" << doc->DocAnchors () << "', "  // List of anchors
            << "doc_hopcount=" << doc->DocHopCount () << " "
            << "where doc_id=" << nDocID;

  s_update3 << "update SERVERS set "
            << "server_type='" << sDocServerType << "' "
            << "where server_id=" << doc->DocServerID ();

  err = OK;
  if (! (err = ExecQuery (s_update1))) 
    if (! (err = ExecQuery (s_update2))) 
      if (nDocStatus == DOC_STAT_OK)
        if (! (err = ExecQuery (s_update3))) 
        {
          String s;
          s << "AddDocInfo() -- update URLS,URLINFO,SERVERS -- "
            << "doc_id=" << nDocID << " "
            << "server_id=" << doc->DocServerID () << ".\n";
          LogLine (s, 2);
          LogLine (s_update3, 2);
        }
        else
        {
          String s;
          s << "AddDocInfo() -- failed3 -- "
            << "doc_id=" << nDocID << "; "
            << "error: '" << Error () << "'.\n";
          LogError (s);
          err = NOTOK;
          //return LogTime (NOTOK, "AddDocInfo", t0);
        }
      else
      {
        String s;
        s << "AddDocInfo() -- update URLS,URLINFO -- "
          << "doc_id=" << nDocID << ".\n";
        LogLine (s, 2);
      }
    else
    {
      String s;
      s << "AddDocInfo() -- failed2 -- "
        << "doc_id=" << nDocID << "; "
        << "error: '" << Error () << "'.\n";
      LogError (s);
 
      err = NOTOK;
      //return LogTime (NOTOK, "AddDocInfo", t0);
    }
  else
  {
    String s;
    s << "AddDocInfo() -- failed1 -- "
      << "doc_id=" << nDocID << "; "
      << "error: '" << Error () << "'.\n"
      << "   query: " << s_update1 << "\n";
    LogError (s);

    err = NOTOK;
    //return LogTime (NOTOK, "AddDocInfo", t0);
  }

  UnLock ();

  return LogTime (err, "AddDocInfo", t0);
}


/********************************************************************/
/*  Update information about a document, only the status, not info  */
/********************************************************************/
int
SQL_db::AddDocInfoLater (int nDocID, DocumentRef *doc, int bUpdateURL)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_update1, s_update2, s_update3;
  int result;
  char **row;

  String sDocURL = doc->DocURL ();
  QuoteString (sDocURL);
  String sDocTitle = doc->DocTitle ();
  QuoteString (sDocTitle);
  String sDocHead = doc->DocHead ();
  QuoteString (sDocHead);
  String sDocServerType = doc->DocServerType ();
  QuoteString (sDocServerType);

  int nDocStatus;
  switch (doc->DocState ())
  {
    case Reference_ok:
    case Reference_new:
      nDocStatus = DOC_STAT_LATER; break;
    case Reference_work:
      nDocStatus = DOC_STAT_WORK; break;
    case Reference_not_found:
      nDocStatus = DOC_STAT_NOT_FOUND; break;
    case Reference_no_host:
      nDocStatus = DOC_STAT_NO_HOST; break;
    case Reference_no_server:
      nDocStatus = DOC_STAT_NO_SERVER; break;
    default:
      nDocStatus = DOC_STAT_UNKNOWN; break;
  }

  String sTable ("URLS write, URLINFO write, SERVERS write");
  Lock (sTable);

  // ... update the docum. in URLS,URLINFO ...
if (bUpdateURL == 1)
  s_update1 << "update URLS set "
            << "doc_url='" << sDocURL << "', "
            << "doc_taccess=from_unixtime(" << doc->DocAccessed () << "), "
            << "doc_status=" << nDocStatus << " "
            << "where doc_id=" << nDocID;
else
  s_update1 << "update URLS set "
            << "doc_taccess=from_unixtime(" << doc->DocAccessed () << "), "
            << "doc_status=" << nDocStatus << " "
            << "where doc_id=" << nDocID;

  s_update3 << "update SERVERS set "
            << "server_type='" << sDocServerType << "' "
            << "where server_id=" << doc->DocServerID ();

  err = OK;
  if (! (err = ExecQuery (s_update1))) 
      if (nDocStatus == DOC_STAT_OK)
        if (! (err = ExecQuery (s_update3))) 
        {
          String s;
          s << "AddDocInfoLater() -- update URLS,SERVERS -- "
            << "doc_id=" << nDocID << " "
            << "server_id=" << doc->DocServerID () << ".\n";
          LogLine (s, 2);
          LogLine (s_update3, 2);
        }
        else
        {
          String s;
          s << "AddDocInfoLater() -- failed3 -- "
            << "doc_id=" << nDocID << "; "
            << "error: '" << Error () << "'.\n";
          LogError (s);
          err = NOTOK;
          //return LogTime (NOTOK, "AddDocInfo", t0);
        }
      else
      {
        String s;
        s << "AddDocInfoLater() -- update URLS,URLINFO -- "
          << "doc_id=" << nDocID << ".\n";
        LogLine (s, 2);
      }
  else
  {
    String s;
    s << "AddDocInfoLater() -- failed1 -- "
      << "doc_id=" << nDocID << "; "
      << "error: '" << Error () << "'.\n"
      << "   query: " << s_update1 << "\n";
    LogError (s);

    err = NOTOK;
    //return LogTime (NOTOK, "AddDocInfo", t0);
  }

  UnLock ();

  return LogTime (err, "AddDocInfo", t0);
}


/*****************************************************************/
/*  Add a word to the LINKS (doc-word)                           */
/*****************************************************************/
int
SQL_db::AddLink (int nDocID, char *pWord, int nWordCount, 
                   int nPosition, int nWeight, int nAnchor)
{
  timeb t0;
  ftime (&t0);

  String sWord (pWord), s_insert;
  int err, nWordID;

  if (AddWord (sWord, nWordID) == NOTOK)
    return LogTime (NOTOK, "AddLink", t0);

  char cNo[3];
  sprintf (cNo, "%02d", nWordID % 100);
  String sNo (cNo);
/*
  s_insert << "insert into LINKS" << sNo << " " 
*/
  s_insert << "insert into TMPLINKS "
           << "(word_id, doc_id, count, position, weight, anchor) "
           << "values (" << nWordID << ","
                         << nDocID << "," 
                         << nWordCount << ","
                         << nPosition << ","
                         << nWeight << ","
                         << nAnchor << ")";

  if (! (err = ExecQuery (s_insert))) 
  {
    String s;
    s << "AddLink() -- insert LINKS" << sNo << " -- "
      << "doc_id=" << nDocID << ", "
      << "word_id=" << nWordID << ", "
      << "word='" << sWord << "'.\n";
    LogLine (s, 2);
  }
  else
  {    
    if (debug)
      { cout << '?'; cout.flush (); }

    String s;
    s << "AddLink() -- failed -- "
      << "doc_id=" << nDocID << ", "
      << "word_id=" << nWordID << ", "
      << "word='" << sWord << "', "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "AddLink", t0);
  }

  return LogTime (OK, "AddLink", t0);
}


/*****************************************************************/
/*  Get all docum.s which containt a specified word              */
/*    - !TODO optimize SQL query for all the words (parser.cc)   */
/*****************************************************************/
int
SQL_db::GetLinks (int nWordID, String &sResult)
{
  timeb t0;
  ftime (&t0);

  int err, i, nLinks = 0;
  int nCount, nDocID, nWeight, nAnchor, nPosition;
  String s_select;
  int result;
  char **row;
/*
  String sWordTxt (sWord);

  if (GetWordID (sWordTxt, nWordID) == NOTOK)
    return LogTime (NOTOK, "GetLinks", t0);
*/

  char cNo[3];
  sprintf (cNo, "%02d", nWordID % 100);
  String sNo (cNo);
  // ... search all documents containing this word ...
  s_select << "select count, doc_id, weight, anchor, position "
           << "from LINKS" << sNo <<" where word_id=" << nWordID;

  if (! (err = ExecQuery (s_select))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      while (row = FetchRow (result))
      {
        nCount    = (row[0])?atoi (row[0]):0;
        nDocID    = (row[1])?atoi (row[1]):0;
        nWeight   = (row[2])?atoi (row[2]):0;
        nAnchor   = (row[3])?atoi (row[3]):0;
        nPosition = (row[4])?atoi (row[4]):0;
        sResult.append ((char *)&nCount,    4); 
        sResult.append ((char *)&nDocID,    4); 
        sResult.append ((char *)&nWeight,   4); 
        sResult.append ((char *)&nAnchor,   4); 
        sResult.append ((char *)&nPosition, 4); 
        //printf ("<%d %d %d %d %d>\n", nCount, nDocID, nWeight, nAnchor, nPosition);
        nLinks++;
      }
      String s;
      s << "GetLinks("
        << "word_id=" << nWordID << ") => "
        << "no.links=" << nLinks << "\n";
      LogLine (s, 2);

      err = OK;
    }
    else
    {
      String s;
      s << "GetLinks("
        << "word_id=" << nWordID << ") => nu such word!\n";
      LogLine (s, 2);

      err = NOTOK;
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetLinks("
      << "word_id=" << nWordID << ") => "
      << "error: '" << Error () << "'!\n";
    LogError (s);

    err = NOTOK;
  }
 
  return LogTime (err, "GetLinks", t0);
}


/*****************************************************************/
/*  Delete everything from LINKS for this document               */
/*****************************************************************/
int
SQL_db::DelLinks (int nDocID)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_delete;

  // ... delete all word links for this docum.
  s_delete << "delete from LINKS where doc_id=" << nDocID;
  
  if (! (err = ExecQuery (s_delete))) 
  {
    String s;
    s << "DelLinks() -- delete LINKS -- "
      << "doc_id=" << nDocID << ".\n";
    LogLine (s, 2);
  }  
  else
  {
    String s;
    s << "DelLinks() -- failed -- "
      << "doc_id=" << nDocID << "; "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "DelLinks", t0);
  }

  return LogTime (OK, "DelLinks", t0);
}

/*****************************************************************/
/*  Delete everything from TMPLINKS                              */
/*****************************************************************/
int
SQL_db::DelTmpLinks (void)
{
  if (strcmp (config["SQL_tmplinks"], "flush"))
    return OK;

  timeb t0;
  ftime (&t0);

  int err;
  String s_delete;

  // ... delete all word links for this docum.
  s_delete << "delete from TMPLINKS";
  
  if (! (err = ExecQuery (s_delete))) 
  {
    String s;
    s << "DelTmpLinks() -- delete TMPLINKS -- \n";
    LogLine (s, 2);
  }  
  else
  {
    String s;
    s << "DelTmpLinks() -- failed -- "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "DelTmpLinks", t0);
  }

  return LogTime (OK, "DelTmpLinks", t0);
}


/*****************************************************************/
/*  Copy from TMPLINKS into LINKSxx                              */
/*****************************************************************/
int
SQL_db::CopyTmpLinks (void)
{
  if (strcmp (config["SQL_tmplinks"], "flush"))
    return OK;

  timeb t0;
  ftime (&t0);

  for (int i = 0; i < 100; i++) {
    if (debug) {
      char s[10];
      sprintf (s, " %2d %%\b\b\b\b\b", i);
      cout << s; cout.flush ();
      //print ("%s", s); fflush (stdout);
    }
    CopyTmpLinksN (i);
  }
  if (debug)
    { cout << " ok  "; cout.flush (); }

  return LogTime (OK, "CopyTmpLinks", t0);
}


/*****************************************************************/
/*  Copy from TMPLINKS into LINKSxx                              */
/*****************************************************************/
int
SQL_db::CopyTmpLinksN (int nNo)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_insert;

  char cNo[3];
  sprintf (cNo, "%02d", nNo);
  String sNo (cNo);
  // ... delete all word links for this docum.
  s_insert << "insert into LINKS" << sNo << " "
           << "select * from TMPLINKS "
           << "where word_id%100=" << nNo;
  
  if (! (err = ExecQuery (s_insert))) 
  {
    String s;
    s << "CopyTmpLinksN(" << nNo << ") -- copy LINKS -- \n";
    LogLine (s, 2);
  }  
  else
  {
    String s;
    s << "CopyTmpLinksN(" << nNo << ") -- failed -- "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "CopyTmpLinksN", t0);
  }

  return LogTime (OK, "CopyTmpLinksN", t0);
}


/*****************************************************************/
/*  Add en email address                                         */
/*****************************************************************/
int
SQL_db::AddEmail (String &sEmail, int nDocID)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_insert, s_select;
  int result;
  char **row;

  QuoteString (sEmail);

  s_insert << "insert into EMAILS (email_addr, doc_id) "
           << "values (" << "'" << sEmail << "',"
                         << nDocID << ")";

  s_select << "select doc_id from EMAILS where email_addr="
           << "'" << sEmail << "'";

  if (! (err = ExecQuery (s_insert))) 
  {
    String s;
    s << "AddEmail() -- insert EMAILS -- "
      << "email='" << sEmail << "'.\n";
    LogLine (s, 2);
  }
  else if (! (err = ExecQuery (s_select)))
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      //printf ("* in %d", atoi (row[0]));
      String s;
      s << "AddEmail() -- select EMAILS -- "
        << "email=" << sEmail << "'.\n";
      LogLine (s, 2);
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "AddEmail() -- insert EMAILS -- "
      << "email='" << sEmail << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "AddEmail", t0);
  }

  return LogTime (OK, "AddEmail", t0);
}


/*********************************************************************/
/*  Add a new WWW server to the SERVERS, it can be a new one or not  */
/*********************************************************************/
int
SQL_db::AddServer (String &sServAddr, int &nServID, int nServStatus)
{
  timeb t0;
  ftime (&t0);

  int err;
  nServID = 0;  
  String s_insert, s_select;
  int result;
  char **row;

  s_insert << "insert into SERVERS (server_addr, server_status) "
           << "values (" << "'" << sServAddr << "',"
                           << nServStatus << ")";

  s_select << "select server_id from SERVERS where server_addr="
           << "'" << sServAddr << "'";

  // insert new server
  if (! (err = ExecQuery (s_insert))) 
  {
    nServID = GetInsertID ();
    String s;
    s << "AddServer() -- insert SERVERS -- "
      << "server=" << sServAddr << "' "
      << "server_id=" << nServID << ".\n";
    LogLine (s, 2);
  }
  // if error check if it already exists
  else if (! (err = ExecQuery (s_select)))
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      // we already have this server, so return the ID
      row = FetchRow (result);
      nServID = (row[0])?atoi (row[0]):0;
      String s;
      s << "AddServer() -- select SERVERS -- "
        << "server='" << sServAddr << "' "
        << "serv_id=" << nServID << ".\n";
      LogLine (s, 2);
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "AddServer() -- failed -- "
      << "server='" << sServAddr << "'; "
      << "error='" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "AddServer", t0);
  }

  return LogTime (OK, "AddServer", t0);
}


/*****************************************************************/
/*  Change the status of the server                              */
/*****************************************************************/
int
SQL_db::UpdateServer (int nServID, int nServStatus)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_update;

  s_update << "update SERVERS set server_status="
           << nServStatus << " where SERVER_ID=" << nServID << ")";

  if (! (err = ExecQuery (s_update))) 
  {
    String s;
    s << "UpdateServer() -- update SERVERS -- "
      << "serv_id=" << nServID << ".\n";
    LogLine (s, 2);
  } 
  else
  {
    String s;
    s << "UpdateServer() -- failed -- "
      << "serv_id=" << nServID << "; "
      << "error='" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "UpdateServer", t0);
  }

  return LogTime (OK, "UpdateServer", t0);
}


/*****************************************************************/
/*  Add a new <word;word_root> pair to WORDS                     */
/*****************************************************************/
int
SQL_db::AddWord (String &sWordTxt, int &nWordID)
{
  timeb t0;
  ftime (&t0);

  String sWordRootTxt = sWordTxt;
  int nWordRootID = 0;

  return AddWord (sWordTxt, nWordID, sWordRootTxt, nWordRootID);
}


int
SQL_db::AddWord (String &sWordTxt, int &nWordID, String &sWordRootTxt, int &nWordRootID)
{
  timeb t0;
  ftime (&t0);

  int err;
  nWordID = 0;
  String s_insert, s_select, s_update;
  int result;
  char **row;

  // ... if don't have the words (cache) hash table ... create it ...
  if (! isWordsHash)
    NewWordsHash ();

  // ... first search the word in the hash table ...
  if (GetWordsHash (sWordTxt, nWordID))
    if (nWordID > 0) {
      if (debug)
        { cout << '.'; cout.flush (); }
      return OK;
    }

// ... if not, then add it to the database also ...
  if (sWordRootTxt.length () > 0)
    s_insert << "insert into WORDS (word_txt, word_lang, word_root_txt, word_root_id) "
             << "values (" << "'" << sWordTxt << "',"
                           << "'xx'," 
                           << "'" << sWordRootTxt << "',"
                           << nWordRootID << ")";
  else
    s_insert << "insert into WORDS (word_txt, word_lang, word_root_txt) "
             << "values (" << "'" << sWordTxt << "',"
                           << "'yy'," 
                           << "'" << sWordRootTxt << "')";

  s_select << "select word_id from WORDS where word_txt="
           << "'" << sWordTxt << "'";

  if (! (err = ExecQuery (s_insert))) 
  {
    if (debug)
      { cout << 'w'; cout.flush (); }
    nWordID = GetInsertID ();

    s_update << "update WORDS set word_root_id=word_id "
             << "where word_id=" << nWordID;

    if (! (err = ExecQuery (s_update))) 
    {
      String s;
      s << "AddWord() -- update WORDS -- "
        << "word='" << sWordTxt << "' "
        << "word_id=" << nWordID << ".\n";
      LogLine (s, 2);
    }
    else 
    {
      String s;
      s << "AddWord() -- failed -- "
        << "word='" << sWordTxt << "' "
        << "word_id=" << nWordID << ".\n";
      LogLine (s, 2);
    }
  }
  else if (! (err = ExecQuery (s_select)))
  {
    if (debug)
      { cout << 'r'; cout.flush (); }

    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      nWordID = (row[0])?atoi (row[0]):0;

      // update the word's root
      if ((sWordRootTxt.length () > 0) && (nWordRootID > 0))
      {
        s_update << "update WORDS set "
                 << "word_root_txt='" << sWordRootTxt << "', "
                 << "word_root_id=" << nWordRootID << " "
                 << "where word_id=" << nWordID;

        if (! (err = ExecQuery (s_update)))
        {
          String s;
          s << "AddWord() -- update WORDS -- "
            << "word='" << sWordTxt << "' "
            << "word_id=" << nWordID << " "
            << "word_root='" << sWordRootTxt << "' "
            << "word_root_id=" << nWordRootID << ".\n";
          LogLine (s, 2);
        }
        else
        {
          String s;
          s << "AddWord() -- failed -- "
            << "word='" << sWordTxt << "' "
            << "word_id=" << nWordID << " "
            << "word_root='" << sWordRootTxt << "' "
            << "word_root_id=" << nWordRootID << "; "
            << "error='" << Error () << "'.\n";
          LogError (s);
        }
      }
      else
      {
        String s;
        s << "AddWord() -- select WORDS -- "
          << "word='" << sWordTxt << "'.\n";
        LogLine (s, 2);
      }
    }
    FreeResult (result);
  }
  else
  {
    if (debug)
      { cout << '?'; cout.flush (); }

    String s;
    s << "AddWord() -- failed -- "
      << "word='" << sWordTxt << "'; "
      << "error='" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "AddWord", t0);
  }

// ... and finaly add the new word to the hash table ...
AddWordsHash (sWordTxt, nWordID);

  return LogTime (OK, "AddWord", t0);
}


int
SQL_db::GetWordID (String &sWord, int &nWordID)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_select;
  int result;
  char **row;

  s_select << "select word_id from WORDS where "
           << "word_txt='" << sWord << "'";

  if (! (err = ExecQuery (s_select))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      nWordID = (row[0])?atoi (row[0]):0;

      String s;
      s << "GetWordID() -- select WORDS -- "
        << "word='" << sWord << "' "
        << "word_id=" << nWordID << "\n";
      LogLine (s, 2);

      err = OK;
    }
    else
    {
      String s;
      s << "GetWordID() -- no such word -- "
        << "word='" << sWord << "'.\n";
      LogLine (s, 2);

      err = NOTOK;
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetWordID() -- failed -- "
      << "word='" << sWord << "'; "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    err = NOTOK;
  }
 
  return LogTime (err, "GetWordID", t0);
}


int
SQL_db::GetWordRoot (String &sWord, int &nWordID, 
                       String &sWordRoot, int &nWordRootID)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_select;
  int result;
  char **row;

  sWordRoot = 0;

  s_select << "select word_root_txt,word_id,word_root_id from WORDS where "
           << "word_txt='" << sWord << "'";

  if (! (err = ExecQuery (s_select))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      row = FetchRow (result);
      if (row[0] && (row[0][0] != 0))
        sWordRoot.append (row[0]);
      else
        sWordRoot = sWord;
      nWordID = (row[1])?atoi (row[1]):0;
      nWordRootID = (row[2])?atoi (row[2]):0;

      String s;
      s << "GetWordRoot("
        << "word='" << sWord << "') => "
        << "word_id=" << nWordID << " "
        << "word_root='" << sWordRoot << "' "
        << "word_root_id=" << nWordRootID << "\n";
      LogLine (s, 2);

      err = OK;
    }
    else
    {
      String s;
      s << "GetWordRoot("
        << "word='" << sWord << "') => no such word!\n";
      LogLine (s, 2);

      err = NOTOK;
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetWordRoot("
      << "word='" << sWord << "') => "
      << "error: '" << Error () << "'!\n";
    LogError (s);

    err = NOTOK;
  }
 
  return LogTime (err, "GetWordRoot", t0);
}


int
SQL_db::ReadWords ()
{
  int err, nWords = 0, nWordID;
  String s_select, sWordTxt;
  int result;
  char **row;

  s_select << "select word_txt,word_id from WORDS order by word_id";

  if (! (err = ExecQuery (s_select))) 
  {
    if (debug)
      { cout << " [read words] "; cout.flush (); }
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      while (row = FetchRow (result))
      {
        sWordTxt = String (row[0]); 
        nWordID = (row[1])?atoi (row[1]):0;
        AddWordsHash (sWordTxt, nWordID);
        nWords++;
      }
      String s;
      s << "ReadWords() => "
        << "no.words=" << nWords << "\n";
      LogLine (s, 2);

      err = OK;
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetWords() => "
      << "error: '" << Error () << "'!\n";
    LogError (s);

    err = NOTOK;
  }
 
  return err;
}


int
SQL_db::GetWords (int nWordRootID, String &sResult, String &sResultID)
{
  timeb t0;
  ftime (&t0);

  int err, nWords = 0;
  String s_select;
  int result;
  char **row;

/* OLD
  s_select << "select word_txt,word_id from WORDS where "
           << "word_root_id=" << nWordRootID << "' "
           << "or (word_root_id=0 and word_txt='" << sWordRoot << "')";
*/
  s_select << "select word_txt,word_id from WORDS where "
           << "word_root_id=" << nWordRootID;

  if (! (err = ExecQuery (s_select))) 
  {
    result = StoreResult ();
    if (NumRows (result) > 0)
    {
      while (row = FetchRow (result))
      {
        sResult.append (row[0]); 
        sResult << " ";
        sResultID.append (row[1]); 
        sResultID << " ";
        nWords++;
      }
      String s;
      s << "GetWords("
        << "word_root_id='" << nWordRootID << "') => "
        << "no.words=" << nWords << " "
        << "words='" << sResult << "' "
        << "wordIDs='" << sResultID << "'\n";
      LogLine (s, 2);

      err = OK;
    }
    else
    {
      String s;
      s << "GetWords("
        << "word_root_id=" << nWordRootID << ") => no such word!\n";
      LogLine (s, 2);

      err = NOTOK;
    }
    FreeResult (result);
  }
  else
  {
    String s;
    s << "GetWords("
      << "word_root_id=" << nWordRootID << ") => "
      << "error: '" << Error () << "'!\n";
    LogError (s);

    err = NOTOK;
  }
 
  return LogTime (err, "GetWords", t0);
}


void
SQL_db::NewWordsHash ()
{
  for (int i = 0; i < 26; i++)
    lWordsHash[i] = new List ();
  ReadWords ();
  isWordsHash = 1;
}


void
SQL_db::AddWordsHash (String &sWordTxt, int nWordID)
{
  char n = sWordTxt[0] - 'a';

  if (SQL_debug > 4)
    printf ("+ %s ... ", sWordTxt.get ());
  fflush (stdout);

  if ((sWordTxt[0] < 'a') || (sWordTxt[0] > 'z'))
  {
    if (SQL_debug > 4)
      printf ("sorry\n");
    return;
  }

  lWordsHash[n]->Add (new HashWord (sWordTxt, nWordID));
  if (SQL_debug > 4)
    printf ("ok\n");
}


int
SQL_db::GetWordsHash (String &sWordTxt, int &nWordID)
{
  HashWord *h;
  String tmp;

  if (SQL_debug > 4)
    printf ("? %s ... ", sWordTxt.get ()); 
  fflush (stdout);

  if ((sWordTxt[0] < 'a') || (sWordTxt[0] > 'z'))
  {
    if (SQL_debug > 4)
      printf ("sorry\n");
    return 0;
  }

  char n = sWordTxt[0] - 'a';
  lWordsHash[n]->Start_Get ();
  while (h = (HashWord *)(lWordsHash[n]->Get_Next ()))
  {
    tmp = h->GetWordTxt ();
//    printf (" %s\n", tmp.get ());
    if (! sWordTxt.compare (&tmp))
    { 
      nWordID = h->GetWordID (); 
      if (SQL_debug > 4)
        printf ("ok\n");
      return 1;
    }
  }

  if (SQL_debug > 4)
    printf ("no\n");
  
  return 0;  // not in the hash
}


int
SQL_db::AddSearch (int &nSearchID, time_t tSearchDate, 
                     String &sWords, int nSearchStatus, int nSearchSecs,
                     String &sRemoteHost, String &sRemoteAddr, 
                     String &sHttpUserAgent, String &sHttpFrom)
{
  timeb t0;
  ftime (&t0);

  int err;
  String s_insert_url;
  int result;
  char **row;

  s_insert_url << "insert into SEARCH "
               << "(search_date, search_words, search_result, search_secs,"
               << " remote_host, remote_addr, http_user_agent, http_from) "
               << "values (" << "from_unixtime(" << tSearchDate << "),"
                             << "'" << sWords << "',"
                             << nSearchStatus << ","
                             << nSearchSecs << ","
                             << "'" << sRemoteHost << "',"
                             << "'" << sRemoteAddr << "',"
                             << "'" << sHttpUserAgent << "',"
                             << "'" << sHttpFrom << "')";

  if (! (err = ExecQuery (s_insert_url)))
  {
    nSearchID = GetInsertID ();

    String s;
    s << "AddSearch() -- insert SEARCH -- ...\n";
    LogLine (s, 2);
  }
  else
  {
    String s;
    s << "AddSearch() -- failed -- "
      << "error: '" << Error () << "'.\n";
    LogError (s);

    return LogTime (NOTOK, "AddSearch", t0);
  }

  return OK;
}


void
SQL_db::QuoteString (String &s)
{
  String ss;

  for (int i = 0; i < s.length (); i++)
    if (s[i] == '\'')
      ss << "\\'";
    else
      ss << s[i];

  s = ss;
}


