//
// SQL_db.h
//
// SQL implementation, written by Zoran Constantinescu <zoran@idi.ntnu.no>
//
#ifndef _SQL_db_h_
#define _SQL_db_h_

#include <Database.h>
#include <DocumentRef.h>

#include <fcntl.h>
#include <stdio.h>
#include <sys/timeb.h>

#define SERV_STAT_OK		0
#define SERV_STAT_NEW		1
#define SERV_STAT_ERR		2
#define SERV_STAT_TIMEOUT	3

#define DOC_STAT_OK		0
#define DOC_STAT_NEW		1
#define DOC_STAT_WORK		2
#define DOC_STAT_UNKNOWN	9
#define DOC_STAT_NOT_FOUND	10
#define DOC_STAT_NO_HOST	11
#define DOC_STAT_NO_SERVER	12
#define DOC_STAT_LATER		13

class SQL_db : public Database
{
    //
    // Construction/Destruction
    //
protected:
    SQL_db();
public:
    ~SQL_db();

    virtual int		OpenReadWrite(char *filename, int mode);
    virtual int		OpenRead(char *filename);
    virtual int		Close();
    virtual int		Delete(const String &) { return 0; }
	
    virtual void	Start_Get() { ; }
    virtual char	*Get_Next() { ; }
    virtual void	Start_Seq(char *str) { ; }
    virtual char	*Get_Next_Seq() { ; }
	
private:
    virtual int		Get(const String &, String &) { return OK; }
    virtual int		Put(const String &, const String &) { return OK; }
    virtual int		Exists(const String &) { return OK; }


/* -- was MySQL-3.0.8b2z */

// specific for each SQL
  protected:
    virtual int	    Connect (char *host, char *db, char *user, char *pswd) = 0;
    virtual int	    Disconnect () = 0;

    virtual int     ExecQuery (String &sQuery) = 0;
    virtual int     GetInsertID () = 0;
    virtual int     StoreResult () = 0;
    virtual void    FreeResult (int result) = 0;
    virtual int     NumRows (int result) = 0;
    virtual char  **FetchRow (int result) = 0;
//    virtual char  **operator [] (int result) = 0;

    virtual char   *Error () = 0;
/*zrn moved to X_db
    virtual void    Lock (String &sTable) = 0;
    virtual void    UnLock () = 0;
*/
    virtual void    QuoteString (String &s);

// generic SQL
  protected:
    void	LogError (String &s);
    void	LogLine (String &s, int level = 0);
    int 	LogTime (int err, char *sFunction, timeb t0);

  public:
    int		AddURL (String &sDocURL, int &nDocID, String &sDocRefs);
    int		GetURL (int nDocID, String &sDocURL);
    int		GetDocID (String &sDocURL, int &nDocID);
    int		GetNewURL (String &sDocURL, int &nDocID, int &nServerID);
    int		GetNewURL (String &sDocURL, int &nDocID, int &nServerID, 
                           String &sLimitDom);
    int		AddDocInfo (int nDocID, DocumentRef *doc, int bUpdateURL);
    int		AddDocInfoLater (int nDocID, DocumentRef *doc, int bUpdateURL);
    int		GetDocInfo (int nDocID, DocumentRef *doc);

    int		AddLink (int nDocID, char *pWord, int nWordCount,
                         int nPosition, int nWeight, int nAnchor);
    int		DelLinks (int nDocID);
    int		GetLinks (int nWordID, String &sResult);
    int		DelTmpLinks (void);
    int		CopyTmpLinks (void);
    int		CopyTmpLinksN (int nNo);

    int		AddEmail (String &sEmail, int nDocID);

    int		AddServer (String &sServURL, int &nServID, int nServStatus);
    int		UpdateServer (int nServID, int nServStatus);

    int		AddWord (String &sWordTxt, int &nWordID);
    int		AddWord (String &sWordTxt, int &nWordID,
                         String &sWordRootTxt, int &nWordRootID);
    int		GetWordRoot (String &sWord, int &nWordID, 
                             String &sWordRoot, int &nWordRootID);
    int		GetWordID (String &sWordRoot, int &nWordID);
    int		GetWords (int nWordRootID, String &sResult, String &sResultID);

    int		AddSearch (int &nSearchID, time_t tSearchDate,
                     String &sWords, int nSearchStatus, int nSearchSecs,
                     String &sRemoteHost, String &sRemoteAddr,
                     String &sHttpUserAgent, String &sHttpFrom);

  private:
    static FILE *SQL_log;    // SQL log file
    static int   SQL_n;      // number of connections
    static int   SQL_id;     // SQL connection ID

    static int   SQL_debug;  // SQL debug level

// hash table stuff
    static int   isWordsHash;
    static List	*lWordsHash[26];

    int 	ReadWords ();
    void	NewWordsHash ();
    void	AddWordsHash (String &sWordTxt, int nWordID);
    int		GetWordsHash (String &sWordTxt, int &nWordID);

};

#endif


